/*--------------------------------------------------------------------------
     "reaction.h" - 'reaction' and library classes (declarations)
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2023 by A. Miyoshi, Hiroshima University
                                       created: Sept. 14, 2002
                                        edited: Aug.  21, 2004
                                        edited: Dec.  26, 2010
                                   last edited: Apr.  17, 2023
--------------------------------------------------------------------------*/

#ifndef _REACTION_H_
#define _REACTION_H_

#include "moleclst.h"
using namespace std;

#define FL_RXNSIT "rxn_site.dat"        // reaction site library
#define FL_RXNTUN "rxn_tune.dat"        // reaction tuning library

/****** 'rxnTunePar' and related classes ************************************/

class rateTuning {
 public:                                         // --- data members ---
  double factA;         // multiplication factor for A
  double shiftE;        // shift of Ea
 public:                                     // --- member functions ---
  rateTuning();
  void clear();
  int restoreFrom(ifstream &ifs);
};

class rxnTunePar : public map<string, rateTuning> {
 public:                                     // --- member functions ---
  rxnTunePar();
  int loadFrom(string flname, libPath &libp);
};

/****** 'rxnLib' and related classes ****************************************/

class reagentInfo {
 public:                                         // --- data members ---
  string rct;
  string prd;
  int isRevrs;
 public:                                     // --- member functions ---
  reagentInfo();
  reagentInfo(string r, string p, int f);
  void clear();
};

class reagentMast : public vector<reagentInfo> {
 public:                                         // --- data members ---
  int numPrior;
 public:                                     // --- member functions ---
  int restoreFrom(ifstream &ifs);
  void print(ostream &os);
};
ostream &operator<<(ostream &os, reagentMast &rm);

class reagentList : public set<string> {
 public:                                     // --- member functions ---
  int restoreFrom(ifstream &ifs);
  void print(ostream &os);
};
ostream &operator<<(ostream &os, reagentList &rg);

class rxnSite {
 public:                                         // --- data members ---
  string snam;
  int prior;
  groupAtom grpa;
  reagentList reag;
  reagentList fullrg;
 public:                                     // --- member functions ---
  int restoreFrom(ifstream &ifs);
  void print(ostream &os);
};
ostream &operator<<(ostream &os, rxnSite &rc);

class rxnSiteList : public vector<rxnSite> {
 public:                                         // --- data members ---
  reagentMast rgMst;
  int priorLim;
 public:                                     // --- member functions ---
  rxnSiteList::iterator find(groupAtom &ga);
  int restoreFrom(ifstream &ifs);
  void print(ostream &os);
};
ostream &operator<<(ostream &os, rxnSiteList &rcl);

class rxnLib : public map<string, rxnSiteList> {
 public:                                     // --- member functions ---
  rxnLib();
  int loadFrom(string flname, libPath &libp);
  void print(ostream &os);
};
ostream &operator<<(ostream &os, rxnLib &rl);


/****** classes related to 'reaction' ***************************************/

// ----- reaction descriptor (rxnDescr) ------------------------------------

class rxnDescr {
 public:                                         // --- data members ---
  vector<string> reac;  // reactants (names)
  vector<string> prod;  // products (names)
  string rate;          // rate macro expression
  double shiftE;        // shift of Ea
  double factA;         // multiplication factor for A
  double shiftB;        // shift of b
  int dupc;             // duplication count
  int isRevrs;          // flag to indicate reversible reaction
  int modTroe;          // modification flag for Troe pars.
  double modAL, modbL, modEaL, modaFc, modT3Fc, modT1Fc, modT2Fc;
 public:                                     // --- member functions ---
  rxnDescr();
  void clear();
  void print(ostream &os);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, rxnDescr &r);

// ----- list of reaction descriptor (rxnList) -----------------------------

class rxnList : public vector<rxnDescr> {
 public:                                     // --- member functions ---
  void appendFrom(rxnList &alst);
  int procDupProd(string pnam);
  int procDupProd(int n, string pnam[]);
  int procDupProdPart(int n, string pnam[]);
  int findDupRevs(string rnam, string pnam);
  void print(ostream &os);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, rxnList &rl);


/****** 'reaction' class - reaction processor *******************************/

class reaction {
 public:                                         // --- data members ---
  rxnList lst;
  rxnList rsvLst;
 private:
  set<string> namPrdGen, namPrdRef;
  int numRxnGen;
  specAdmin *spAd;
  groupLib *grLb;
  rxnLib *rxLb;
  rxnTunePar *rxTp;
  typeMolec *tyMol;
  outStrms *oStrm;
  vector<molecule> *fuels;
  molecList radFuel;
  molecList ro2Fuel;
  // - for debugging
  // int tmpFlag;
  //
 public:                                     // --- member functions ---
  void clear();
  void initPrivMemb(specAdmin &sa, groupLib &g, rxnLib &rl, rxnTunePar &rt,
   typeMolec &tm, outStrms &os, vector<molecule> &f);
  void reportGenStat();
  int reactQTop();
 private:

  // ===== general purpose private functions ===========================
  //
  void prepRadRo2Fuel();
  int checkIsFuel(molecule &m);
  int checkIsRadFuel(molecule &m);
  int checkIsRo2Fuel(molecule &m);
  int checkIsRxnDone(molecule &m);
  void procProd(molecule &prod, string bnam);
  void procProdAutoBase(molecule &prod);
  void outLstClear();
  void outLstResrv();
  void endReact(molecule &m);

  // ===== type-specific reaction processors ===========================
  //
  void rxnHabsFuel(molecule &m);
  void rxnHabsPart(molecule &m);
  void rxnHabsNewRO2(molecule &ro2);
  void rxnHabsAlkene(molecule &m);
  void rxnHabsCycEther(molecule &m);
  void rxnHabsKetAld(molecule &m);
  void rxnHabsAldOnly(molecule &m);
  void rxnOabsNewR(molecule &r);
  void rxnOabsNewRO2(molecule &ro2);
  void rxnDecmpNonRad(molecule &m);
  void rxnDecmpCHnonR(molecule &m);
  void rxnDecmpRad(molecule &m);
  void rxnDecmpCycEther(molecule &m);
  // void rxnDecmpSpec(molecule &m, string bspc);
  void rxnDecmpQaOOH(molecule &m);
  void rxnDecmpQOOH(molecule &m);
  // void rxnDecmpKetOOH(molecule &m); // - renamed Sept. 28, 2010
  void rxnDecmpROOH(molecule &m, double ofsNHA);
  void rxnDecmpKetAld(molecule &m);
  void rxnDecmpRCO(molecule &m);
  void rxnDecmpOCCV(molecule &m);
  void rxnDecmpVinoxy(molecule &m);
  void rxnDecmpKetR(molecule &m);
  void rxnIsomRad(molecule &m);
  void rxnIsomOOQOOH(molecule &m);
  void rxnRecmbQOOHO2(molecule &m);
  void rxnRecmbRadO2(molecule &m);
  void rxnElimHO2RO2(molecule &m);
  void rxnRO2RO2(molecule &ro2);
  void rxnAddRadAlkene(molecule &m);
  void rxnAddHelAlkene(molecule &m);  // - added 230417
  void rxnOaddAlkene(molecule &m);
  void rxnO2addAllyl(molecule &m);
  void rxnO2addVinyl(molecule &m);
  void rxnO2addHORalp(molecule &m);
  void rxnO2addVinoxy(molecule &m);
  void rxnO2addKetyl(molecule &m);
  void rxnO2addKetR(molecule &m);
  void rxnKetene(molecule &m);
  void rxnIntAddRdO2(molecule &m);
  void rxnEnol(molecule &m);

  // ===== molecular manipulation functions ============================
  //
  int molDissoc(molecule &m, int id1, int id2, molecule &p1, molecule &p2,
   int isRad);
  int molCycDissoc(molecule &m, int id1, int id2, molecule &prd);
  int molCycRadDis(molecule &m, int id1, int id2, molecule &prd);
  void molDissocHalf(molecule &m, int cid, int idx, molecule &prd,
   int isRad);
  void followConnxn(molecInfo &mi, int cid, int idx, map<int, int> &l);
  int molDisKetR(molecule &m, molecule &p1, molecule &p2);
  void molAddRcO2(molecule &m, molecule &prd);
  void molAddRcO(molecule &m, molecule &prd);
  void molRmvRcO(molecule &m, molecule &prd);
  void molAddOH(molecule &m, molecule &prd, int id);
  void molAddHO2(molecule &m, molecule &prd, int id);
  int rfmAddOat(molecule &tm, int idr);
  int rfmBridgeOatDB(molecule &tm, int id1, int id2);
};

#endif

